from enum import Enum
from inspect import getmodule
from logging import FileHandler, getLogger, StreamHandler
from pathlib import Path
from signal import SIGUSR1
from sys import stdout
from traceback import StackSummary
from typing import Optional
import argparse
import configparser
import datetime as dt
import logging
import ssl

from trio import open_memory_channel, open_nursery, open_signal_receiver

try:
	from trio.lowlevel import current_root_task, Task
except ImportError:
	from trio.hazmat import current_root_task, Task

import trio_asyncio
import triopg

from strikebot import count_tracker_impl
from strikebot.db import Client, Messenger
from strikebot.live_ws import HealingReadPool, PoolMerger
from strikebot.reddit_api import ApiClientPool


_DEBUG_LOG_PATH: Optional[Path] = None  # file to write debug logs to, if any


_TaskKind = Enum(
	"_TaskKind",
	[
		"API_WORKER",
		"DB_CLIENT",
		"TOKEN_UPDATE",
		"TRACKER",
		"WS_MERGE_READER",
		"WS_MERGE_TIMER",
		"WS_REFRESHER",
		"WS_WORKER",
	]
)


def _classify_task(task: Task) -> _TaskKind:
	mod_name = getmodule(task.coro.cr_code).__name__
	if mod_name.startswith("trio_asyncio."):
		return None
	elif task.coro.__name__ == "_signal_handler_impl":
		return None
	else:
		by_coro_name = {
			"db_client_impl": _TaskKind.DB_CLIENT,
			"token_updater_impl": _TaskKind.TOKEN_UPDATE,
			"event_reader_impl": _TaskKind.WS_MERGE_READER,
			"timeout_handler_impl": _TaskKind.WS_MERGE_TIMER,
			"conn_refresher_impl": _TaskKind.WS_REFRESHER,
			"count_tracker_impl": _TaskKind.TRACKER,
			"_reader_impl": _TaskKind.WS_WORKER,
			"worker_impl": _TaskKind.API_WORKER,
		}
		return by_coro_name[task.coro.__name__]


def _get_all_tasks():
	[ta_loop] = [
		task
		for nursery in current_root_task().child_nurseries
		for task in nursery.child_tasks
		if getmodule(task.coro.cr_code).__name__.startswith("trio_asyncio.")
	]
	for nursery in ta_loop.child_nurseries:
		yield from nursery.child_tasks


def _print_task_info():
	groups = {kind: [] for kind in _TaskKind}
	for task in _get_all_tasks():
		kind = _classify_task(task)
		if kind:
			coro = task.coro
			frame_tuples = []
			while coro is not None:
				if hasattr(coro, "cr_frame"):
					frame_tuples.append((coro.cr_frame, coro.cr_frame.f_lineno))
					coro = coro.cr_await
				else:
					frame_tuples.append((coro.gi_frame, coro.gi_frame.f_lineno))
					coro = coro.gi_yieldfrom
			groups[kind].append(StackSummary.extract(iter(frame_tuples)))
	for kind in _TaskKind:
		print(kind.name)
		for ss in groups[kind]:
			print("  task")
			for line in ss.format():
				print("    " + line, end = "")


async def _signal_handler_impl():
	with open_signal_receiver(SIGUSR1) as sig_src:
		async for _ in sig_src:
			_print_task_info()


ap = argparse.ArgumentParser(__package__)
ap.add_argument("config_path")
args = ap.parse_args()


parser = configparser.ConfigParser()
with open(args.config_path) as config_file:
	parser.read_file(config_file)

main_cfg = parser["config"]

api_pool_error_delay = dt.timedelta(seconds = main_cfg.getfloat("API pool error delay"))
api_pool_error_window = dt.timedelta(seconds = main_cfg.getfloat("API pool error window"))
auth_ids = set(map(int, main_cfg["auth IDs"].split()))
bot_user = main_cfg["bot user"]
enforcing = main_cfg.getboolean("enforcing", fallback = True)

raw_log_level = main_cfg.get("log level", "WARNING")
try:
	log_level = int(raw_log_level)
except ValueError:
	log_level = raw_log_level

read_only = main_cfg.getboolean("read-only", fallback = False)
reorder_buffer_time = dt.timedelta(seconds = main_cfg.getfloat("reorder buffer time"))
request_queue_limit = main_cfg.getint("request queue limit")
thread_id = main_cfg["thread ID"]
update_retention = dt.timedelta(seconds = main_cfg.getfloat("update retention time"))
ws_parity_time = dt.timedelta(seconds = main_cfg.getfloat("WS parity time"))
ws_pool_size = main_cfg.getint("WS pool size")
ws_silent_limit = dt.timedelta(seconds = main_cfg.getfloat("WS silent limit"))
ws_warmup = dt.timedelta(seconds = main_cfg.getfloat("WS warmup time"))


db_cfg = parser["db connect params"]
cert_path = db_cfg.pop("sslcert", None)
key_path = db_cfg.pop("sslkey", None)
key_pass = db_cfg.pop("sslpassword", None)

getters = {
	"port": db_cfg.getint,
}
db_connect_kwargs = {k: getters.get(k, db_cfg.get)(k) for k in db_cfg}

if cert_path:
	# patch TLS client cert auth support; asyncpg adds it in a later version
	db_ssl_ctx = ssl.create_default_context()
	db_ssl_ctx.load_cert_chain(
		cert_path,
		key_path,
		None if key_pass is None else lambda: key_pass,
	)
	db_connect_kwargs["ssl"] = db_ssl_ctx


if read_only and enforcing:
	raise RuntimeError("can't use read-only mode with enforcing on")


logger = getLogger(__package__)
logger.setLevel(logging.NOTSET - 1)  # NOTSET means inherit from parent; we use handlers to filter

handler = StreamHandler(stdout)
handler.setLevel(log_level)
handler.setFormatter(logging.Formatter("{asctime:23}: {name:17} {levelname:8}: {message}", style = "{"))
logger.addHandler(handler)

if _DEBUG_LOG_PATH:  # TODO remove this ad hoc setup
	debug_handler = FileHandler(_DEBUG_LOG_PATH, mode = "w")
	debug_handler.setLevel(logging.DEBUG)
	debug_handler.setFormatter(logging.Formatter("{asctime:23}: {name:17} {levelname:8}: {message}", style = "{"))
	logger.addHandler(debug_handler)


async def main():
	async with (
		triopg.connect(**db_connect_kwargs) as db_conn,
		open_nursery() as nursery_a,
		open_nursery() as nursery_b,
		open_nursery() as ws_pool_nursery,
		open_nursery() as nursery_c,
	):
		nursery_a.start_soon(_signal_handler_impl)

		client = Client(db_conn)
		db_messenger = Messenger(client)
		nursery_a.start_soon(db_messenger.db_client_impl)

		api_pool = ApiClientPool(
			auth_ids, db_messenger, request_queue_limit, api_pool_error_window, api_pool_error_delay,
			logger.getChild("api")
		)
		nursery_a.start_soon(api_pool.token_updater_impl)
		for _ in auth_ids:
			await nursery_a.start(api_pool.worker_impl)

		(message_tx, message_rx) = open_memory_channel(0)
		(pool_event_tx, pool_event_rx) = open_memory_channel(0)

		merger = PoolMerger(pool_event_rx, message_tx, ws_parity_time, ws_warmup, logger.getChild("merge"))
		nursery_b.start_soon(merger.event_reader_impl)
		nursery_b.start_soon(merger.timeout_handler_impl)

		ws_pool = HealingReadPool(
			ws_pool_nursery, ws_pool_size, thread_id, api_pool, pool_event_tx, logger.getChild("live_ws"),
			ws_silent_limit
		)
		nursery_c.start_soon(ws_pool.conn_refresher_impl)

		nursery_c.start_soon(
			count_tracker_impl, message_rx, api_pool, reorder_buffer_time, thread_id, bot_user, enforcing, read_only,
			update_retention, logger.getChild("track")
		)
		await ws_pool.init_workers()

trio_asyncio.run(main)
