from collections import deque
from contextlib import suppress
from dataclasses import dataclass
from typing import Any, AsyncContextManager, Optional
import datetime as dt
import json
import logging
import math

from trio_websocket import ConnectionClosed, HandshakeError, open_websocket_url, WebSocketConnection
import trio

from strikebot.common import int_digest, obj_digest
from strikebot.reddit_api import AboutLiveThreadRequest


@dataclass
class _PoolEvent:
	timestamp: float  # Trio time


@dataclass
class _Message(_PoolEvent):
	data: Any
	scope: Any

	def dedup_tag(self):
		"""A hashable essence of the contents of this message."""
		if self.data["type"] == "update":
			return ("update", self.data["payload"]["data"]["id"])
		elif self.data["type"] in {"strike", "delete"}:
			return (self.data["type"], self.data["payload"])
		else:
			raise ValueError(self.data["type"])


@dataclass
class _ConnectionDown(_PoolEvent):
	scope: trio.CancelScope


@dataclass
class _ConnectionUp(_PoolEvent):
	scope: trio.CancelScope


class HealingReadPool:
	def __init__(
		self, pool_nursery: trio.Nursery, size, live_thread_id, api_client_pool, pool_event_tx, logger: logging.Logger,
		silent_limit: dt.timedelta
	):
		assert size >= 2
		self._nursery = pool_nursery
		self._size = size
		self._live_thread_id = live_thread_id
		self._api_client_pool = api_client_pool
		self._pool_event_tx = pool_event_tx
		self._logger = logger
		self._silent_limit = silent_limit

		(self._refresh_queue_tx, self._refresh_queue_rx) = trio.open_memory_channel(size)

		# number of workers who have stopped receiving updates but whose tasks aren't yet stopped
		self._closing = 0

	async def init_workers(self):
		for _ in range(self._size):
			await self._spawn_reader()
		if not self._nursery.child_tasks:
			raise RuntimeError("Unable to create any WS connections")

	async def _reader_impl(self, conn_ctx: AsyncContextManager, cancel_scope, refresh_tx):
		# TODO could use task's implicit cancel scope
		try:
			async with conn_ctx as conn:
				self._logger.debug("scope up: {}".format(obj_digest(cancel_scope)))
				async with refresh_tx:
					silent_timeout = False
					with cancel_scope, suppress(ConnectionClosed):
						while True:
							with trio.move_on_after(self._silent_limit.total_seconds()) as timeout_scope:
								message = await conn.get_message()

							if timeout_scope.cancelled_caught:
								if len(self._nursery.child_tasks) - self._closing == self._size:
									silent_timeout = True
									break
								else:
									self._logger.debug("not replacing connection {}; {} tasks, {} closing".format(
										obj_digest(cancel_scope),
										len(self._nursery.child_tasks),
										self._closing,
									))
							else:
								event = _Message(trio.current_time(), json.loads(message), cancel_scope)
								await self._pool_event_tx.send(event)

					self._closing += 1
					if silent_timeout:
						await conn.aclose()
						self._logger.debug("replacing WS connection {} due to silent timeout".format(obj_digest(cancel_scope)))
					elif cancel_scope.cancelled_caught:
						await conn.aclose(1008, "Server unexpectedly stopped sending messages")
						self._logger.warning("replacing WS connection {} due to missed update".format(obj_digest(cancel_scope)))
					else:
						self._logger.warning("replacing WS connection {} closed by server".format(obj_digest(cancel_scope)))

					refresh_tx.send_nowait(cancel_scope)
					self._logger.debug("scope down: {} ({} tasks, {} closing)".format(
						obj_digest(cancel_scope),
						len(self._nursery.child_tasks),
						self._closing,
					))
					self._closing -= 1
		except HandshakeError:
			self._logger.error("handshake error while opening WS connection")

	async def _new_connection(self) -> Optional[AsyncContextManager[WebSocketConnection]]:
		request = AboutLiveThreadRequest(self._live_thread_id)
		resp = await self._api_client_pool.make_request(request)
		if resp.status_code == 200:
			url = json.loads(resp.text)["data"]["websocket_url"]
			return open_websocket_url(url)
		else:
			self._logger.error(f"thread info request failed (HTTP {resp.status_code}); can't open WS connection")
			return None

	async def _spawn_reader(self) -> None:
		conn = await self._new_connection()
		if conn:
			new_scope = trio.CancelScope()
			await self._pool_event_tx.send(_ConnectionUp(trio.current_time(), new_scope))
			refresh_tx = self._refresh_queue_tx.clone()
			self._nursery.start_soon(self._reader_impl, conn, new_scope, refresh_tx)

	async def conn_refresher_impl(self):
		"""Task to monitor and replace WS connections as they disconnect or go silent."""
		async with self._refresh_queue_rx:
			async for old_scope in self._refresh_queue_rx:
				await self._pool_event_tx.send(_ConnectionDown(trio.current_time(), old_scope))
				await self._spawn_reader()
				if not self._nursery.child_tasks:
					raise RuntimeError("WS pool depleted")


def _tag_digest(tag):
	return int_digest(hash(tag))


def _format_scope_list(scopes):
	return ", ".join(sorted(map(obj_digest, scopes)))


class PoolMerger:
	@dataclass
	class _Bucket:
		start: float
		recipients: set

	def __init__(self, pool_event_rx, message_tx, parity_timeout: dt.timedelta, conn_warmup: dt.timedelta, logger: logging.Logger):
		"""
		:param parity_timeout: max delay between message receipt on different connections
		:param conn_warmup: max interval after connection within which missed updates are allowed
		"""
		self._pool_event_rx = pool_event_rx
		self._message_tx = message_tx
		self._parity_timeout = parity_timeout
		self._conn_warmup = conn_warmup
		self._logger = logger

		self._buckets = {}
		self._scope_activations = {}
		self._pending = deque()
		self._outgoing_scopes = set()
		(self._timer_poke_tx, self._timer_poke_rx) = trio.open_memory_channel(math.inf)

	def _log_current_buckets(self):
		self._logger.debug("current buckets: {}".format(", ".join(map(_tag_digest, self._buckets.keys()))))

	async def event_reader_impl(self):
		"""Drop unused messages, deduplicate useful ones, and communicate with the timeout handler."""
		async with self._pool_event_rx, self._timer_poke_tx:
			async for event in self._pool_event_rx:
				if isinstance(event, _ConnectionUp):
					# An early add of an active scope could mean it's expected on a message that fired before it opened,
					# resulting in a false positive for replacement. The timer task merges these in among the buckets by
					# timestamp to avoid that.
					self._pending.append(event)
				elif isinstance(event, _Message):
					if event.data["type"] in ["update", "strike", "delete"]:
						tag = event.dedup_tag()
						self._logger.debug("recv {} from {}".format(_tag_digest(tag), obj_digest(event.scope)))
						if tag in self._buckets:
							b = self._buckets[tag]
							b.recipients.add(event.scope)
							# If this scope is the last one for this bucket we could clear the bucket here, but since
							# connections sometimes get repeat messages and second copies can arrive before the first
							# copy has arrived on all connections, leaving the bucket open to absorb repeats can reduce
							# the likelihood of a second bucket being allocated late for the second copy of a message
							# and causing unnecessary connection replacement.
						elif event.scope not in self._outgoing_scopes:
							sane = (
								event.scope in self._scope_activations
								or any(e.scope == event.scope for e in self._pending)
							)
							if sane:
								self._logger.debug("new bucket {}: {}".format(_tag_digest(tag), tag))
								self._buckets[tag] = self._Bucket(event.timestamp, {event.scope})
								self._log_current_buckets()
								await self._message_tx.send(event)
							else:
								raise RuntimeError("recieved message from unrecognized WS connection")
					else:
						self._logger.debug("recv type {!r} from {} (discarding)".format(event.data["type"], obj_digest(event.scope)))
				elif isinstance(event, _ConnectionDown):
					# We don't need to worry about canceling this scope at all, so no need to require it for parity for
					# any message, even older ones.  The scope may be gone already, if we canceled it previously.
					self._scope_activations.pop(event.scope, None)
					self._pending = deque(ev for ev in self._pending if ev.scope is not event.scope)
					self._outgoing_scopes.discard(event.scope)
				else:
					raise TypeError(f"Expected pool event, found {event!r}")
				self._timer_poke_tx.send_nowait(None)  # may have new work for the timer

	async def timeout_handler_impl(self):
		"""When connections have had enough time to reach parity on a message, signal replacement of any slackers."""
		async with self._timer_poke_rx:
			while True:
				if self._buckets:
					now = trio.current_time()
					(tag, bucket) = min(self._buckets.items(), key = lambda i: i[1].start)

					# Make sure the right scope set is ready for the next bucket up
					while self._pending and self._pending[0].timestamp < bucket.start:
						ev = self._pending.popleft()
						self._scope_activations[ev.scope] = ev.timestamp

					target_scopes = {
						scope for (scope, active) in self._scope_activations.items()
						if active + self._conn_warmup.total_seconds() < now
					}
					if now > bucket.start + self._parity_timeout.total_seconds():
						filled = target_scopes & bucket.recipients
						missing = target_scopes - bucket.recipients
						extra = bucket.recipients - target_scopes

						self._logger.debug("expiring bucket {}".format(_tag_digest(tag)))
						if filled:
							self._logger.debug("  filled {}: {}".format(len(filled), _format_scope_list(filled)))
						if missing:
							self._logger.debug("  missing {}: {}".format(len(missing), _format_scope_list(missing)))
						if extra:
							self._logger.debug("  extra {}: {}".format(len(extra), _format_scope_list(extra)))

						for scope in missing:
							self._outgoing_scopes.add(scope)
							scope.cancel()
						del self._buckets[tag]
						self._log_current_buckets()
					else:
						await trio.sleep_until(bucket.start + self._parity_timeout.total_seconds())
				else:
					try:
						await self._timer_poke_rx.receive()
					except trio.EndOfChannel:
						break
