"""Unbounded blocking queues for Trio"""

from collections import deque
from dataclasses import dataclass
from functools import total_ordering
from typing import Any, Iterable
import heapq

try:
	from trio.lowlevel import ParkingLot
except ImportError:
	from trio.hazmat import ParkingLot


class Queue:
	def __init__(self):
		self._deque = deque()
		self._empty_wait = ParkingLot()

	def push(self, el: Any) -> None:
		self._deque.append(el)
		self._empty_wait.unpark()

	def extend(self, els: Iterable[Any]) -> None:
		for el in els:
			self.push(el)

	async def pop(self) -> Any:
		if not self._deque:
			await self._empty_wait.park()
		return self._deque.popleft()

	def __len__(self):
		return len(self._deque)


@total_ordering
@dataclass
class _ReverseOrdWrapper:
	inner: Any

	def __lt__(self, other):
		return other.inner < self.inner


class MaxHeap:
	def __init__(self):
		self._heap = []
		self._empty_wait = ParkingLot()

	def push(self, item):
		heapq.heappush(self._heap, _ReverseOrdWrapper(item))
		self._empty_wait.unpark()

	async def pop(self):
		if not self._heap:
			await self._empty_wait.park()
		return heapq.heappop(self._heap).inner

	def __len__(self) -> int:
		return len(self._heap)
