"""Multi-auth HTTP connection pool with Reddit API wrappers and rate limiting."""

from abc import ABCMeta, abstractmethod
from dataclasses import dataclass, field
from functools import total_ordering
from socket import EAI_AGAIN, EAI_FAIL, gaierror
import datetime as dt
import logging

from asks.response_objects import Response
import asks
import trio

from strikebot import __version__ as VERSION
from strikebot.common import obj_digest
from strikebot.queue import MaxHeap, Queue


REQUEST_DELAY = dt.timedelta(seconds = 1)

TOKEN_UPDATE_DELAY = dt.timedelta(minutes = 15)

USER_AGENT_FMT = f"any:net.jcornell.strikebot.{{auth_id}}:v{VERSION} (by /u/jaxklax)"

API_BASE_URL = "https://oauth.reddit.com"


@total_ordering
class _Request(metaclass = ABCMeta):
	_SUBTYPE_PRECEDENCE = None  # assigned later

	def __lt__(self, other):
		if type(self) is type(other):
			return self._subtype_cmp_key() < other._subtype_cmp_key()
		else:
			prec = self._SUBTYPE_PRECEDENCE
			return prec.index(type(other)) < prec.index(type(self))

	def __eq__(self, other):
		if type(self) is type(other):
			return self._subtype_cmp_key() == other._subtype_cmp_key()
		else:
			return False

	@abstractmethod
	def _subtype_cmp_key(self):
		# a large key corresponds to a high priority
		raise NotImplementedError()

	@abstractmethod
	def to_asks_kwargs(self):
		raise NotImplementedError()


@dataclass(eq = False)
class StrikeRequest(_Request):
	thread_id: str
	update_name: str
	update_ts: dt.datetime

	def to_asks_kwargs(self):
		return {
			"method": "POST",
			"path": f"/api/live/{self.thread_id}/strike_update",
			"data": {
				"api_type": "json",
				"id": self.update_name,
			},
		}

	def _subtype_cmp_key(self):
		return self.update_ts


@dataclass(eq = False)
class DeleteRequest(_Request):
	thread_id: str
	update_name: str
	update_ts: dt.datetime

	def to_asks_kwargs(self):
		return {
			"method": "POST",
			"path": f"/api/live/{self.thread_id}/delete_update",
			"data": {
				"api_type": "json",
				"id": self.update_name,
			},
		}

	def _subtype_cmp_key(self):
		return -self.update_ts.timestamp()


@dataclass(eq = False)
class AboutLiveThreadRequest(_Request):
	thread_id: str
	created: float = field(default_factory = trio.current_time)

	def to_asks_kwargs(self):
		return {
			"method": "GET",
			"path": f"/live/{self.thread_id}/about",
			"params": {
				"raw_json": "1",
			},
		}

	def _subtype_cmp_key(self):
		return -self.created


@dataclass(eq = False)
class _BaseUpdatePostRequest(_Request, metaclass = ABCMeta):
	thread_id: str
	body: str
	created: float = field(default_factory = trio.current_time)

	def to_asks_kwargs(self):
		return {
			"method": "POST",
			"path": f"/live/{self.thread_id}/update",
			"data": {
				"api_type": "json",
				"body": self.body,
			},
		}


@dataclass(eq = False)
class ReportUpdateRequest(_BaseUpdatePostRequest):
	def _subtype_cmp_key(self):
		return -self.created


@dataclass(eq = False)
class CorrectionUpdateRequest(_BaseUpdatePostRequest):
	def _subtype_cmp_key(self):
		return -self.created


# highest priority first
_Request._SUBTYPE_PRECEDENCE = [
	AboutLiveThreadRequest,
	StrikeRequest,
	CorrectionUpdateRequest,
	ReportUpdateRequest,
	DeleteRequest,
]


@dataclass
class AppCooldown:
	auth_id: int
	ready_at: float  # Trio time


class ApiClientPool:
	def __init__(
		self,
		auth_ids: set[int],
		db_messenger: "strikebot.db.Messenger",
		request_queue_limit: int,
		error_window: dt.timedelta,
		error_delay: dt.timedelta,
		logger: logging.Logger,
	):
		self._auth_ids = auth_ids
		self._db_messenger = db_messenger
		self._request_queue_limit = request_queue_limit
		self._error_window = error_window
		self._error_delay = error_delay
		self._logger = logger

		now = trio.current_time()
		self._tokens = {}
		self._app_queue = Queue()
		self._waiting = {id_: AppCooldown(id_, now) for id_ in auth_ids}
		self._request_queue = MaxHeap()

		# pool-wide API error backoff
		self._last_error = None
		self._global_resume = None

		self._session = asks.Session(connections = len(auth_ids))
		self._session.base_location = API_BASE_URL

	async def _update_tokens(self):
		tokens = await self._db_messenger.do("get_auth_tokens", (self._auth_ids,))
		self._tokens.update(tokens)

		awaken_auths = self._waiting.keys() & tokens.keys()
		self._app_queue.extend(self._waiting.pop(auth_id) for auth_id in awaken_auths)
		self._logger.debug("updated API tokens")

	async def token_updater_impl(self):
		last_update = None
		while True:
			if last_update is not None:
				await trio.sleep_until(last_update + TOKEN_UPDATE_DELAY.total_seconds())

			if last_update is None:
				last_update = trio.current_time()
			else:
				last_update += TOKEN_UPDATE_DELAY.total_seconds()

			await self._update_tokens()

	def _check_queue_size(self) -> None:
		if len(self._request_queue) > self._request_queue_limit:
			raise RuntimeError("request queue size exceeded limit")

	async def make_request(self, request: _Request) -> Response:
		(resp_tx, resp_rx) = trio.open_memory_channel(0)
		self.enqueue_request(request, resp_tx)
		async with resp_rx:
			return await resp_rx.receive()

	def enqueue_request(self, request: _Request, resp_tx = None) -> None:
		self._logger.debug("request {}: {}".format(obj_digest(request), type(request).__name__))
		self._request_queue.push((request, resp_tx))
		self._check_queue_size()
		self._logger.debug(f"{len(self._request_queue)} requests in queue")

	async def worker_impl(self, task_status):
		task_status.started()
		while True:
			(request, resp_tx) = await self._request_queue.pop()
			cooldown = await self._app_queue.pop()
			await trio.sleep_until(cooldown.ready_at)
			if self._global_resume:
				await trio.sleep_until(self._global_resume)

			asks_kwargs = request.to_asks_kwargs()
			headers = asks_kwargs.setdefault("headers", {})
			headers.update({
				"Authorization": "Bearer {}".format(self._tokens[cooldown.auth_id]),
				"User-Agent": USER_AGENT_FMT.format(auth_id = cooldown.auth_id),
			})

			request_time = trio.current_time()
			try:
				resp = await self._session.request(**asks_kwargs)
			except gaierror as e:
				if e.errno in [EAI_FAIL, EAI_AGAIN]:
					# DNS failure, probably temporary
					error = True
				else:
					raise
			else:
				resp.body  # read response
				error = False
				wait_for_token = False
				log_suffix = " (request {})".format(obj_digest(request))
				if resp.status_code == 429:
					# We disagreed about the rate limit state; just try again later.
					self._logger.warning("rate limited by Reddit API" + log_suffix)
					error = True
				elif resp.status_code == 401:
					self._logger.warning("got HTTP 401 from Reddit API" + log_suffix)
					error = True
					wait_for_token = True
				elif resp.status_code in [404, 500, 503]:
					self._logger.warning(f"got HTTP {resp.status_code} from Reddit API, retrying" + log_suffix)
					error = True
				elif 400 <= resp.status_code < 500:
					# If we're doing something wrong, let's catch it right away.
					raise RuntimeError(f"unexpected client error response: {resp.status_code}" + log_suffix)
				else:
					if resp.status_code != 200:
						raise RuntimeError(f"unexpected status code {resp.status_code}")
					self._logger.debug("success" + log_suffix)
					if resp_tx:
						await resp_tx.send(resp)

			if error:
				self._request_queue.push((request, resp_tx))
				self._check_queue_size()
				if self._last_error:
					spread = dt.timedelta(seconds = request_time - self._last_error)
					if spread <= self._error_window:
						self._global_resume = request_time + self._error_delay.total_seconds()
				self._last_error = request_time

			cooldown.ready_at = request_time + REQUEST_DELAY.total_seconds()
			if wait_for_token:
				self._waiting[cooldown.auth_id] = cooldown
				if not self._app_queue:
					self._logger.error("all workers waiting for API tokens")
			else:
				self._app_queue.push(cooldown)
